@@ -2,7 +2,7 @@
 #
 # This code is part of lattpy.
 #
-# Copyright (c) 2021, Dylan Jones
+# Copyright (c) 2022, Dylan Jones
 #
 # This code is licensed under the MIT License. The copyright notice in the
 # LICENSE file in the root directory and this permission notice shall
@@ -23,6 +23,7 @@ from .utils import (
 
 from .spatial import (
     distance,
+    distances,
     interweave,
     vindices,
     vrange,
@@ -55,7 +56,7 @@ del get_versions
 def simple_chain(a=1.0, atom=None, neighbors=1):
     latt = Lattice.chain(a)
     latt.add_atom(atom=atom)
-    latt.set_num_neighbors(neighbors)
+    latt.add_connections(neighbors)
     return latt
 
 
@@ -63,7 +64,7 @@ def alternating_chain(a=1.0, atom1=None, atom2=None, x0=0.0, neighbors=1):
     latt = Lattice.chain(a)
     latt.add_atom(pos=(0.0 + x0) * a, atom=atom1)
     latt.add_atom(pos=(0.5 + x0) * a, atom=atom2)
-    latt.set_num_neighbors(neighbors)
+    latt.add_connections(neighbors)
     return latt
 
 
@@ -74,30 +75,31 @@ def alternating_chain(a=1.0, atom1=None, atom2=None, x0=0.0, neighbors=1):
 def simple_square(a=1.0, atom=None, neighbors=1):
     latt = Lattice.square(a)
     latt.add_atom(atom=atom)
-    latt.set_num_neighbors(neighbors)
+    latt.add_connections(neighbors)
     return latt
 
 
 def simple_rectangular(a1=1.0, a2=1.0, atom=None, neighbors=1):
     latt = Lattice.rectangular(a1, a2)
     latt.add_atom(atom=atom)
-    latt.set_num_neighbors(neighbors)
+    latt.add_connections(neighbors)
     return latt
 
 
 def centered_rectangular(a1=1.0, a2=1.0, atom=None, neighbors=1):
-    latt = Lattice([[a1, 0], [a1/2, a2/2]])
+    latt = Lattice([[a1, 0], [a1 / 2, a2 / 2]])
     latt.add_atom(atom=atom)
-    latt.set_num_neighbors(neighbors)
+    latt.add_connections(neighbors)
     return latt
 
 
-def graphene(a=1.0, neighbors=1):
-    at = Atom("C")
+def graphene(a=1.0):
+    at1 = Atom("C1")
+    at2 = Atom("C2")
     latt = Lattice.hexagonal(a)
-    latt.add_atom([0, 0], at)
-    latt.add_atom([a, 0], at)
-    latt.set_num_neighbors(neighbors)
+    latt.add_atom([0, 0], at1)
+    latt.add_atom([a, 0], at2)
+    latt.add_connection(at1, at2, analyze=True)
     return latt
 
 
@@ -108,13 +110,13 @@ def graphene(a=1.0, neighbors=1):
 def simple_cubic(a=1.0, atom=None, neighbors=1):
     latt = Lattice.sc(a)
     latt.add_atom(atom=atom)
-    latt.set_num_neighbors(neighbors)
+    latt.add_connections(neighbors)
     return latt
 
 
 def nacl_structure(a=1.0, atom1="Na", atom2="Cl", neighbors=1):
     latt = Lattice.fcc(a)
     latt.add_atom(pos=[0, 0, 0], atom=atom1)
-    latt.add_atom(pos=[a/2, a/2, a/2], atom=atom2)
-    latt.set_num_neighbors(neighbors)
+    latt.add_atom(pos=[a / 2, a / 2, a / 2], atom=atom2)
+    latt.add_connections(neighbors)
     return latt
